/*
 * kylin-status-manager
 *
 * Copyright (C) 2022, KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Authors: wangguixi <wangguixi@kylinos.cn>
 *
 */
#include <QGuiApplication>
#include <QDebug>
#include <QDBusConnection>
#include <QDBusInterface>
#include <QDBusReply>
#include "autoswitch.h"

// 状态管理器
#define SETTINGS_KYLINSTATUSMANAGER    "org.kylin-status-manager"
#define MODE_MONITER         "modemonitor" // 是否监听角度值
#define TABLET_MODE          "tablet-mode"
#define AUTO_ROTATION        "auto-rotation" // 旋转
#define ROTATIONS            "rotations"
#define KEYBOARD_ENABLE      "keyboardenable"
// 主题
#define THEME_QT_SCHEMA "org.ukui.style"
#define THEME_Style_Name "styleName"

#define MODE_TYPE       1
#define ROTATIONS_TYPE   0
#define KEYBOARD_TYPE     2

autoSwitch::autoSwitch(QObject *parent) :
    QObject(parent)
{
    qInfo()<<"==="<<__FILE__<<__LINE__<<" initialization===";
    //注册DBus
    regist_dbus();
    //初始化本地数据gsetting （自动旋转、平板pc、自动切换）
    QByteArray id(SETTINGS_KYLINSTATUSMANAGER);
    if(QGSettings::isSchemaInstalled(id)) {
        status_manager_Setting=new QGSettings(id);
    } else {
        qInfo()<<__FILE__<<__LINE__<<"org.kylin-status-manager is not install";
        return;
    }

    //主题 （监听主题的gsetting,对外提供dbus接口信号，注：主题没有对外提供此方法，所以由状态管理器提供给搜狗）
    QByteArray id3(THEME_QT_SCHEMA);
    if(QGSettings::isSchemaInstalled(id3))
    {
        ukui_style_Setting = new QGSettings(id3);
        // 监听当前主题变化
        connect(ukui_style_Setting, &QGSettings::changed, [this](const QString & key) {
            qInfo()<<__FILE__<<__LINE__<<"org.ukui.style changed"<<key;
            if (key == THEME_Style_Name) {
                emit stylename_change_signal(ukui_style_Setting->get(THEME_Style_Name).toString());
            }
        });
    } else {
        qInfo()<<__FILE__<<__LINE__<<"org.ukui.style is not install";
    }


    /*设备初始化init()*/

    devicestatus = new Devices_Status_Monitoring;
    init();

    connect(devicestatus,&Devices_Status_Monitoring::keyboard_enable_signal,
            this,&autoSwitch::processing_keyboard_enable_signal);
    connect(devicestatus,&Devices_Status_Monitoring::rotations_change_signal,
            this,&autoSwitch::processing_rotations_change_signal);
    connect(devicestatus,&Devices_Status_Monitoring::mode_change_signal,
            this,&autoSwitch::processing_mode_change_signal);

    //旋转翻折延时防抖处理
//    debounce_timer= new QTimer();
//    connect(debounce_timer, &QTimer::timeout, this, &autoSwitch::mode_debounce_timeout);
    debounce_timer1= new QTimer();
    connect(debounce_timer1, &QTimer::timeout, this, &autoSwitch::rotations_debounce_timeout);
    debounce_timer2= new QTimer();
    connect(debounce_timer2, &QTimer::timeout, this, &autoSwitch::keyboard_debounce_timeout);

    //该槽函数是为了在通过侧边栏切换时键盘状态不能更新的情况设置的。
    connect(this,&autoSwitch::mode_change_signal,[this](){
        keyboard_debounce(100);
    });
    qInfo()<<"==="<<__FILE__<<__LINE__<<" initialization end===";
}
autoSwitch::~autoSwitch()
{
    if(debounce_timer) {
        delete debounce_timer;
    }
    if(debounce_timer1) {
        delete debounce_timer1;
    }
    if(debounce_timer2) {
        delete debounce_timer2;
    }

    if(ukui_style_Setting) {
        delete ukui_style_Setting;
    }
    if(status_manager_Setting) {
        delete status_manager_Setting;
    }

    delete devicestatus;
}
void autoSwitch::init()
{
    qDebug()<<__FILE__<<__LINE__<<"====start of initialization====";
    if(status_manager_Setting) {
        ksmstruct.modemonitor = status_manager_Setting->get(MODE_MONITER).toBool();//模式监控开关
        ksmstruct.autorotation =status_manager_Setting->get(AUTO_ROTATION).toBool();//旋转监控开关
        ksmstruct.tablet_mode=status_manager_Setting->get(TABLET_MODE).toBool();//机器模式状态
        ksmstruct.rotation=status_manager_Setting->get(ROTATIONS).toString();//旋转状态
        ksmstruct.keyboardenable=status_manager_Setting->get(KEYBOARD_ENABLE).toBool();
    } else {
        qInfo()<<__FILE__<<__LINE__<<"init fail";
        return;
    }
    //初始化 是否开始自动切换
    if (ksmstruct.modemonitor == true) {
        //根据键盘情况更新模式
        bool tablet_mode= devicestatus->get_tablet_mode_status();
        if (tablet_mode == false) {
            //pc模式
            set_tabletmode(false,"ksm","init");
            set_rotation("normal","ksm","init");
            set_keyboardenable(true,"ksm","init");

        } else {
            //pad模式
            set_tabletmode(true,"ksm","init");
            set_keyboardenable(false,"ksm","init");
            if(ksmstruct.autorotation == true) {
                update_rotation_status();
            } else {//没有开全自动旋转是根据上一次的记录还是根据默认情况设置normal??
                set_rotation("normal","ksm","init");//根据默认设置
//                set_rotation("ksmstruct.rotation","ksm","init");//根据上一次设置
            }
        }
    } else {
        //恢复上次状态
        set_tabletmode(ksmstruct.tablet_mode,"ksm","init");
        bool keyboardstatus= devicestatus->get_keyboard_enable_result();
        set_keyboardenable(keyboardstatus,"ksm","init");
        if(ksmstruct.tablet_mode == false) {
            //pc
            set_rotation("normal","ksm","init");
        } else {
            //pad
            if(ksmstruct.autorotation == true) {
                update_rotation_status();
            } else {//没有开全自动旋转是根据上一次的记录还是根据默认情况设置normal??
                set_rotation("normal","ksm","init");//根据默认设置
//                set_rotation("ksmstruct.rotation","ksm","init");//根据上一次设置
            }
        }
    }
    qDebug()<<__FILE__<<__LINE__<<"====End of initialization====";
}
//注册dbus
void autoSwitch::regist_dbus()
{
    qDebug()<<__FILE__<<__LINE__<<"====RegistDbus====";
    //被拉起后的回调
    QDBusInterface Sessioninterface("org.gonme.SessionManager",
                             "/org/gonme/SessionManager",
                             "org.gonme.SessionManager",
                                QDBusConnection::sessionBus());

    //调用session的接口，通知管理服务已经启动
    QDBusReply<QString> reply = Sessioninterface.call("startupfinished","kylin-status-manager","startupfinished");

    QDBusConnection  sessionBus= QDBusConnection::sessionBus();
    if (sessionBus.registerService("com.kylin.statusmanager.interface")) {
        qDebug() <<"Regist Dbus Service success";
        sessionBus.registerObject("/", this, QDBusConnection :: ExportAllSlots | QDBusConnection :: ExportAllSignals);
    } else {
        qCritical() << "Register DBus Service Error:" << sessionBus.lastError().message();
    }
    qDebug()<<__FILE__<<__LINE__<<"====End of RegistDbus====";
}
// 处理系列
void autoSwitch::processing_mode_change_signal(bool tablet_mode)
{
    qInfo()<<__FILE__<<__LINE__<<"processing_mode_change_signal :"<<tablet_mode;
    update_mode_status(tablet_mode);
}
void autoSwitch::processing_keyboard_enable_signal(bool enable)
{
    qInfo()<<__FILE__<<__LINE__<<"processing_keyboard_enable_signal :"<<enable;
    keyboard_debounce(100);
}
void autoSwitch::processing_rotations_change_signal(QString rotations)
{
    qInfo()<<__FILE__<<__LINE__<<"processing_rotations_change_signal"<<rotations;
    //判断当前旋转状态与新接收的信号的旋转状态是否一致
    if(ksmstruct.rotation == rotations) {
        return ;
    }
    rotations_debounce(100);
}
void autoSwitch::mode_debounce_timeout() //模式切换防抖
{
//    ++foldtime;
//    if (foldtime == 1) {
//        foldtime = 0;
//        debounce_timer->stop();
//        //得到最新的状态进行设置
//        update_mode_status();
//    }
}
void autoSwitch::rotations_debounce_timeout() //旋转防抖
{
    ++rotetime;
    if (rotetime == 7) {
        rotetime = 0;
        debounce_timer1->stop();
        //得到最新的状态进行设置
        update_rotation_status();
    }
}
void autoSwitch::keyboard_debounce_timeout() //键盘缓冲
{
    ++keyboardtim;
    if (keyboardtim == 14) {
        keyboardtim = 0;
        debounce_timer2->stop();
        //得到最新的状态进行设置
        update_keyboard_status();
    }
}
void autoSwitch::mode_debounce(int msec)
{
    if(debounce_timer->isActive()){
        debounce_timer->stop();
        foldtime=0;
        debounce_timer->start(msec);
    }else{
        debounce_timer->start(msec);
    }
}
void autoSwitch::rotations_debounce(int msec)
{
    if(debounce_timer1->isActive()){
        debounce_timer1->stop();
        keyboardtim=0;
        debounce_timer1->start(msec);
    }else{
        debounce_timer1->start(msec);
    }
}
void autoSwitch::keyboard_debounce(int msec)
{
    if(debounce_timer2->isActive()){
        debounce_timer2->stop();
        rotetime=0;
        debounce_timer2->start(msec);
    }else{
        debounce_timer2->start(msec);
    }
}
//update 系列  不考虑判断，由其他函数判断完成后，调用
//根据当前键盘状态 更新当前键盘情况  不考虑判断
void autoSwitch::update_keyboard_status()
{
    qDebug()<<"update_keyboard_status";
    bool keyboardenable= devicestatus->get_keyboard_enable_result();
    set_keyboardenable(keyboardenable,"ksm","update");
}
//根据当前的gsetting更新缓存结构体数据
void autoSwitch::update_ksmstruct()
{
    if(status_manager_Setting) {
        ksmstruct.modemonitor = status_manager_Setting->get(MODE_MONITER).toBool();//模式监控开关
        ksmstruct.autorotation = status_manager_Setting->get(AUTO_ROTATION).toBool();//旋转监控开关
        ksmstruct.tablet_mode = status_manager_Setting->get(TABLET_MODE).toBool();//机器模式状态
        ksmstruct.rotation = status_manager_Setting->get(ROTATIONS).toString();//旋转状态
        ksmstruct.keyboardenable = status_manager_Setting->get(KEYBOARD_ENABLE).toBool();

    }
}
//更新模式。并根据现在的设置，更新其他的状态，无缓冲
void autoSwitch::update_mode_status(bool tablet_mode)
{
    qDebug()<<"update_mode_status"<< tablet_mode;
    if (!tablet_mode) {
        //pc模式
        set_tabletmode(false,"ksm","update");
        set_rotation("normal","ksm","update");
    } else {
        //pad模式
        set_tabletmode(true,"ksm","update");
        update_rotation_status();
    }
}
//更新旋转状态，并根据现在的设置，判断是否需要旋转，pc下紧张旋转
void autoSwitch::update_rotation_status()
{
    qDebug()<<"update_rotation_status";
    if(ksmstruct.tablet_mode){
        QString rotationstatus=devicestatus->get_rotation_status();
//        if(ksmstruct.rotation != rotationstatus){
           set_rotation(rotationstatus,"ksm","update");
//        }
    }
}
//get 系列
int autoSwitch::get_inputmethod_mode()//获得当前键盘状态情况
{ 
    qDebug()<<"get_inputmethod_mode";
    if(status_manager_Setting) {
        if(status_manager_Setting->get(KEYBOARD_ENABLE).toBool()) {
            return 1;
        }else {
            return 0;
        }
    }else {
        return 1;
    }
}
QString autoSwitch::get_current_stylename()//获取当前主体名称 //不用改
{
    if(ukui_style_Setting) {
        return ukui_style_Setting->get(THEME_Style_Name).toString();
    }else {
        return NULL;
    }
}
bool autoSwitch::get_current_tabletmode()//获得当前的模式
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(TABLET_MODE).toBool();
    }else {
        return 0;
    }
}
QString autoSwitch::get_current_rotation()//获得当前旋转情况
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(ROTATIONS).toString();
    }else {
        return "normal";
    }
}
bool autoSwitch::get_modemonitor()//获取自动切换模式开关情况
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(MODE_MONITER).toBool();
    }else {
        return false;
    }
}
bool autoSwitch::get_auto_rotation()//获取自动旋转开关情况
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(AUTO_ROTATION).toBool();
    }else {
        return false;
    }
}
/*设置接口
 * 修改采用gsetting的方式进行设置，gsetting文件仅仅用来记录
 * 通过函数调用的方式进行更改状态，增加函数复用
*/
bool autoSwitch::set_rotation(QString rotation,QString who,QString why)
{
    qInfo()<<__FILE__<<__LINE__<<who<<"set_rotation = "<<rotation<<"to"<<why;
    if(rotation!="normal"&&rotation!="upside-down"&&rotation!="left"&& rotation!="right")
    {
        return 0;
    }
    if(status_manager_Setting) {
      status_manager_Setting->set(ROTATIONS,rotation);
      Q_EMIT rotations_change_signal(rotation);
      update_ksmstruct();
      return 1;
    } else {
        return 0;
    }
}
bool autoSwitch::set_tabletmode(bool tablet_mode,QString who,QString why)
{
    qInfo()<<__FILE__<<__LINE__<<who<<"set_tabletmode = "<<tablet_mode<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(TABLET_MODE,tablet_mode);
        Q_EMIT mode_change_signal(tablet_mode);
        if(tablet_mode)//pad
        {
            if(ksmstruct.autorotation){
                QString rotationstatus=devicestatus->get_rotation_status();
                set_rotation(rotationstatus,"ksm","update");
            }
        }else {
            if(ksmstruct.rotation!="normal")
            {
                status_manager_Setting->set(ROTATIONS,"normal");
                Q_EMIT rotations_change_signal("normal");
            }
        }
        update_ksmstruct();
        return 1;
    } else {
        return 0;
    }
}
bool autoSwitch::set_modemonitor(bool modemonitor,QString who,QString why)
{
    qInfo()<<__FILE__<<__LINE__<<who<<"set_modemonitor = "<<modemonitor<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(MODE_MONITER,modemonitor);
        Q_EMIT modemonitor_change_signal(modemonitor);
        if(modemonitor) {
            bool tablet_mode=devicestatus->get_tablet_mode_status();
            update_mode_status(tablet_mode);
        }
        update_ksmstruct();
        return 1;
    }else {
        return 0;
    }
}
bool autoSwitch::set_auto_rotation(bool autorotation,QString who,QString why)
{
    qInfo()<<__FILE__<<__LINE__<<who<<"set_auto_rotation = "<<autorotation<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(AUTO_ROTATION,autorotation);
        Q_EMIT auto_rotation_change_signal(autorotation);
        if(autorotation) {
            update_rotation_status();
        }
        update_ksmstruct();
        return 1;
    }else {
        return 0;
    }
}

bool autoSwitch::set_keyboardenable(bool keyboardenable,QString who,QString why)
{
    qInfo()<<__FILE__<<__LINE__<<who<<"set_keyboardenable = "<<keyboardenable<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(KEYBOARD_ENABLE,keyboardenable);
        Q_EMIT inputmethod_change_signal(keyboardenable);
        update_ksmstruct();
        return 1;
    } else {
        return 0;
    }
}
//QString autoSwitch::get_current_machine_category()
//{
//    setting->beginGroup("machinecategory");
//    QString machine_category= setting->value("machine_category").toString();
//    setting->endGroup();
//    setting->sync();
//    return machine_category;
//}

